#ifndef HPMODEL_H
#define HPMODEL_H

#include <vector>
#include <string>
#include <stdexcept>
#include <iostream>
#include <map>
#include <set>

enum DIRECTION {FORWARD, UP, DOWN, LEFT, RIGHT};
enum ABSOLUTE_DIRECTION {NORTH, EAST, SOUTH, WEST, ABS_UP, ABS_DOWN};
enum ROTATION {NO_ROTATION, ROTATE_90, ROTATE_180, ROTATE_270};



class Point {
public:
    Point();

    int x, y, z;
    inline bool operator==(const Point& other) const;
    inline bool operator<(const Point& other) const;
    Point(int x, int y, int z);
};

class ScorePair {
public:
    ScorePair(Point p1, Point p2);
    Point p1, p2;
    bool operator==(const ScorePair& other) const;
    bool operator<(const ScorePair& other) const;
};

class Acid {
public:
    int x, y, z;
    bool hydrophobic;
    Acid(int x, int y, int z, bool hydrophobic);
    Acid();
    inline bool operator==(const Acid& other) const;
    inline bool operator<(const Acid& other) const;
    void print() const;
    // "Don't you DARE discard it >:(" -CLion
    [[nodiscard]] bool coords_match(Acid other) const;
    Point getPoint();
private:
    Point point;
};

class HPModel {
private:
    std::string sequence;
    std::vector<Acid> acids;
    std::map<Point, Acid> acidMap{};
    std::set<ScorePair> scoringPairs;
    ABSOLUTE_DIRECTION currentDirection;
    ABSOLUTE_DIRECTION previousDirection;
    ROTATION rotation;
    Acid getHead();
    static std::string directionToString(DIRECTION direction);
    bool isInsertValid(Acid toInsert);
    bool isCurrentAcidHydrophobic();
    static ABSOLUTE_DIRECTION getPlaneDirectionChange(ABSOLUTE_DIRECTION absolute_direction, DIRECTION relative_direction);
    ABSOLUTE_DIRECTION getUpDownDirectionAndUpdatePrevious(DIRECTION relative_direction);
    void updateCurrentDirection(DIRECTION direction);
    void addPoint(DIRECTION direction, bool isHydrophobic);
    void rotate(int amount);
    DIRECTION getRotatedRelativeDirection(DIRECTION relative_direction);
    static ABSOLUTE_DIRECTION invertDirection(ABSOLUTE_DIRECTION direction);
    void insert(Acid acid);
    void updateScore(Acid latestAcid);
    std::vector<Acid> getNearbyPotentialAcids(Acid currentAcid, Acid connectingAcid);
    void doFolds(const std::string& folds);
public:
    std::vector<Acid> getPoints();
    void fold(DIRECTION direction);
    HPModel(const std::string& sequence);
    HPModel(const std::string& sequence, const std::string& folds);
    HPModel clone();
    void print() const;
    std::string pointsToJson() const;
    /**
     * Score is guaranteed to be up to date with current model configuration
     * @return The score :D
     */
    int getScore();
};

#endif
